/* This file is part of the LWRPC library */
#ifndef _LWRPC_DATA_INCLUDED
#define _LWRPC_DATA_INCLUDED

#include "uth.h"
#include "utils.h"

struct lwrpc_data {
	unsigned int		_execution_mode;
	void			(*_entry_routine)(int,char **);
	unsigned int		_thread_stacksize;
	unsigned int		_global_vps;
	unsigned int		_kthreads;
	unsigned int		_physcpus;
	int			_mpi_rank;
	int			_mpi_nodes;
	unsigned int		_internode_stealing;
	unsigned int		_appl_finished;
	unsigned int		_tobe_finished;
	unsigned int		_barrier_arrived;
	unsigned int		_thread_safe;
	unsigned int		_non_blocking;
	unsigned int		_block_idle_threads;
	unsigned int		_old_block_idle_threads;
	unsigned int		_package_suspended;

	unsigned int		_master_noworker;

	pthread_t		_server_thread;
	pthread_t		_worker_thread[MAX_CPUS];
	lwrpc_t			*_scheduler_rpc[MAX_CPUS];
	lwrpc_t			*_yield_rpc[MAX_CPUS];	

	lwrpc_t			*_cleanup_rpc[MAX_CPUS][CACHE_LINE_SIZE/sizeof(lwrpc_t *)];

	queue_t			_reuse_q;				
	queue_t			_private_grq;                      
	queue_t			_public_grq;                      
	queue_t			_public_lrq[MAX_CPUS];             

	unsigned long		_created[MAX_CPUS];
	unsigned long		_executed[MAX_CPUS];

	unsigned long		_created2[MAX_CPUS];
	unsigned long		_executed2[MAX_CPUS];

	_lock_t			_datalock;				
	_lock_t			_reductionlock;				
	_lock_t			_commlock;

};

extern struct lwrpc_data *lwrpc_data;

#define entry_routine		lwrpc_data->_entry_routine

#define execution_mode		lwrpc_data->_execution_mode

#define thread_stacksize	lwrpc_data->_thread_stacksize
#define global_vps		lwrpc_data->_global_vps
#define kthreads		lwrpc_data->_kthreads
#define physcpus		lwrpc_data->_physcpus
#define appl_finished		lwrpc_data->_appl_finished
#define tobe_finished		lwrpc_data->_tobe_finished
#define barrier_arrived		lwrpc_data->_barrier_arrived
#define thread_safe		lwrpc_data->_thread_safe
#define non_blocking		lwrpc_data->_non_blocking
#define block_idle_threads	lwrpc_data->_block_idle_threads
#define	old_block_idle_threads	lwrpc_data->_old_block_idle_threads
#define package_suspended	lwrpc_data->_package_suspended

#define master_noworker		lwrpc_data->_master_noworker
#define worker_thread		lwrpc_data->_worker_thread
#define private_grq		lwrpc_data->_private_grq
#define public_grq		lwrpc_data->_public_grq
#define public_lrq		lwrpc_data->_public_lrq
#define reuse_q			lwrpc_data->_reuse_q
#define datalock		lwrpc_data->_datalock
#define scheduler_rpc		lwrpc_data->_scheduler_rpc
#define yield_rpc		lwrpc_data->_yield_rpc
#define cleanup_rpc		lwrpc_data->_cleanup_rpc
#define reductionlock		lwrpc_data->_reductionlock
#define commlock		lwrpc_data->_commlock
#define internode_stealing	lwrpc_data->_internode_stealing
#define server_thread		lwrpc_data->_server_thread
#define mpi_rank		lwrpc_data->_mpi_rank
#define mpi_nodes		lwrpc_data->_mpi_nodes

#define created			lwrpc_data->_created
#define executed		lwrpc_data->_executed

#define created2		lwrpc_data->_created2
#define executed2		lwrpc_data->_executed2

#endif




