/* This file is part of the LWRPC library */
#include <lwrpc_internal.h>
#include <lwrpc.h>
#include <stdlib.h>
#include <string.h>

void rq_init ()
{
	int i;

    	_queue_init(&reuse_q);
	_queue_init (&private_grq);
	_queue_init (&public_grq);
	for (i = 0; i < MAX_CPUS; i++) {
		_queue_init (&public_lrq[i]);
	}
}

static void lwrpc_to_i_reuseq (lwrpc_t *rpc)
{
	_enqueue_head(&reuse_q, rpc);
}

static void lwrpc_to_i_reuseq_end (lwrpc_t *rpc)
{
	_enqueue_tail(&reuse_q, rpc);
}

static lwrpc_t *lwrpc_i_reuseq_dequeue ()
{
	lwrpc_t *rpc = NULL;

	_dequeue(&reuse_q, &rpc);
	return rpc;
}

lwrpc_t *_lwrpc_get_reused_desc()
{
        lwrpc_t *rpc = NULL;
	char *ptr;
	static unsigned long offset = sizeof(_lock_t);
	static unsigned long lwrpc_size = sizeof(lwrpc_t);

        rpc = lwrpc_i_reuseq_dequeue();
	if (rpc != NULL) {
		ptr = (char *) rpc;
		ptr += offset;
		memset(ptr, 0, lwrpc_size - offset);
	}
	else {
		rpc = calloc(1, sizeof(lwrpc_t));
	}

        return rpc;
}

void _lwrpc_put_reused_desc(lwrpc_t *rpc)
{
        lwrpc_to_i_reuseq_end(rpc);
}

void lwrpc_to_i_pq (lwrpc_t *rpc)
{
	_enqueue_head(&private_grq, rpc);
	thread_signal();
}

void lwrpc_to_i_pq_end (lwrpc_t *rpc)
{
	_enqueue_tail(&private_grq, rpc);
	thread_signal();
}

lwrpc_t *lwrpc_i_pq_dequeue ()
{
	lwrpc_t *rpc = NULL;

	_dequeue(&private_grq, &rpc);
	return rpc;
}

void lwrpc_to_i_rq (lwrpc_t *rpc)
{
	_enqueue_head(&public_grq, rpc);
	thread_signal();
}

void lwrpc_to_i_rq_end (lwrpc_t *rpc)
{
	_enqueue_tail(&public_grq, rpc);
	thread_signal();
}

lwrpc_t *lwrpc_i_rq_dequeue ()
{
	lwrpc_t *rpc = NULL;

	_dequeue(&public_grq, &rpc);
	return rpc;
}

void lwrpc_to_i_lrq (int qid, lwrpc_t *rpc)
{
	_enqueue_head(&public_lrq[qid], rpc);
	thread_signal();
}

void lwrpc_to_i_lrq_end (int qid, lwrpc_t *rpc)
{
	_enqueue_tail(&public_lrq[qid], rpc);
	thread_signal();
}

lwrpc_t *lwrpc_i_lrq_dequeue (int qid)
{
	lwrpc_t *rpc = NULL;

	_dequeue(&public_lrq[qid], &rpc);
	return rpc;
}

static void read_arguments(lwrpc_t *rpc)
{
	int i;
	for (i = 0; i < rpc->narg; i++)	rpc->temparg[i] = rpc->localarg[i];
}

void lwrpc_to_nrq (int target_node, lwrpc_t *rpc)
{
	rpc->insert_in_front = 1;	
	rpc->inter_node = 1;
	rpc->target_queue = -1;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		read_arguments(rpc);
		lwrpc_to_i_rq(rpc);
	}
}

void lwrpc_to_nrq_end (int target_node, lwrpc_t *rpc)
{
	rpc->inter_node = 1;
	rpc->target_queue = -1;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		read_arguments(rpc);
		lwrpc_to_i_rq_end(rpc);
	}
}

void lwrpc_to_rq (lwrpc_t *rpc)
{
	static int initialized = 0;
	static int target_node;
	int total_nodes = lwrpc_num_nodes();

	if (initialized == 0) {
		initialized = 1;
		target_node = lwrpc_node_id();
	}

	rpc->insert_in_front = 1;
	rpc->insert_private = 0;
	rpc->inter_node = 1;
	rpc->target_queue = -1;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		read_arguments(rpc);
		lwrpc_to_i_rq(rpc);

	}
	target_node = (target_node+1) % total_nodes;
}

void lwrpc_to_rq_end__ (lwrpc_t *rpc)
{
	int i;
	static int initialized = 0;
	static int target_node;
	int total_nodes = lwrpc_num_nodes();

	if (initialized == 0) {
		initialized = 1;
		target_node = lwrpc_node_id();
	}

	rpc->insert_private = 0;
	rpc->inter_node = 1;
	rpc->target_queue = -1;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		for (i = 0; i < rpc->narg; i++)	rpc->temparg[i] = rpc->localarg[i];
		lwrpc_to_i_rq_end(rpc);
	}
	target_node = (target_node+1) % total_nodes;
}


void lwrpc_to_rq_end (lwrpc_t *rpc)
{
	int i;
	static int initialized = 0;
	static int target_worker;
	int total_workers = lwrpc_num_workers();
	int target_node, target_queue;

	if (initialized == 0) {
		initialized = 1;
		target_worker = lwrpc_worker_id();
	}

	if (lwrpc_num_nodes() == 1) {
		lwrpc_to_i_rq_end(rpc);
		return;
	}

	target_node  = global_thread_id_to_node_id(target_worker);
	target_queue = global_thread_id_to_local_thread_id(target_worker);

	rpc->insert_private = 0;
	rpc->inter_node = 1;
	rpc->target_queue = target_queue;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		for (i = 0; i < rpc->narg; i++)	rpc->temparg[i] = rpc->localarg[i];
		lwrpc_to_i_rq_end(rpc);
	}
	target_worker = (target_worker+1) % total_workers;
}

void lwrpc_to_lrq (int target, lwrpc_t *rpc)
{
	int target_node = global_thread_id_to_node_id(target);
	int target_queue = global_thread_id_to_local_thread_id(target);

	rpc->insert_in_front = 1;
	rpc->inter_node = 1;
	rpc->target_queue = target_queue;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		read_arguments(rpc);
		lwrpc_to_i_lrq(rpc->target_queue, rpc);
	}
}

void lwrpc_to_lrq_end (int target, lwrpc_t *rpc)
{
	int target_node, target_queue;
	
	if (lwrpc_num_nodes() == 1) {
		lwrpc_to_i_lrq_end(target, rpc);
		return;
	}

	target_node  = global_thread_id_to_node_id(target);
	target_queue = global_thread_id_to_local_thread_id(target);

	rpc->inter_node = 1;
	rpc->target_queue = target_queue;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		read_arguments(rpc);
		lwrpc_to_i_lrq_end(rpc->target_queue, rpc);
	}
}

void lwrpc_to_npq (int target_node, lwrpc_t *rpc)
{
	rpc->insert_private = 1;
	rpc->insert_in_front = 1;
	rpc->inter_node = 1;

	rpc->target_queue = -1;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		read_arguments(rpc);
		lwrpc_to_i_pq(rpc);
	}
}

void lwrpc_to_npq_end (int target_node, lwrpc_t *rpc)
{
	rpc->insert_private = 1;
	rpc->inter_node = 1;
	rpc->target_queue = -1;
	if (lwrpc_node_id() != target_node) {
		send_descriptor(target_node, rpc, LWRPC_NORMAL_ENQUEUE);
		_lwrpc_put_reused_desc(rpc);
	}
	else {
		read_arguments(rpc);
		lwrpc_to_i_pq_end(rpc);
	}
}
